package net.bytebuddy.test.packaging;

import net.bytebuddy.description.method.AbstractMethodDescriptionTest;

public abstract class MethodDescriptionTestHelper extends AbstractMethodDescriptionTest.PublicType {
    /* empty */
}
