package net.bytebuddy.test.packaging;

import net.bytebuddy.description.field.AbstractFieldDescriptionTest;

public class FieldDescriptionTestHelper extends AbstractFieldDescriptionTest.PublicType {
    /* empty */
}
