package net.bytebuddy.test.packaging;

public class SimpleOptionalType {

    private SimpleType simpleType;

    private static final String FOO = "foo";

    public String foo() {
        return FOO;
    }
}
