/*
 * Copyright 2014 - Present Rafael Winterhalter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
/**
 * All classes and types in this package are related to creating a {@link net.bytebuddy.dynamic.DynamicType} by
 * creating a subclass of a given type.
 */
@NeverNull.ByDefault
package net.bytebuddy.dynamic.scaffold.subclass;

import net.bytebuddy.utility.nullability.NeverNull;
