/*
 * Copyright 2014 - Present Rafael Winterhalter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.bytebuddy.implementation.bytecode;

import net.bytebuddy.implementation.Implementation;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;

/**
 * A stack manipulation that negates a number on the operand stack.
 */
public enum Negation implements StackManipulation {

    /**
     * Negates an integer or an integer-compatible value.
     */
    INTEGER(Opcodes.INEG),

    /**
     * Negates a long.
     */
    LONG(Opcodes.LNEG),

    /**
     * Negates a float.
     */
    FLOAT(Opcodes.FNEG),

    /**
     * Negates a double.
     */
    DOUBLE(Opcodes.DNEG);

    /**
     * The opcode to apply.
     */
    private final int opcode;

    /**
     * Creates a new negation.
     *
     * @param opcode The opcode to apply.
     */
    Negation(int opcode) {
        this.opcode = opcode;
    }

    /**
     * {@inheritDoc}
     */
    public boolean isValid() {
        return true;
    }

    /**
     * {@inheritDoc}
     */
    public Size apply(MethodVisitor methodVisitor, Implementation.Context implementationContext) {
        methodVisitor.visitInsn(opcode);
        return StackManipulation.Size.ZERO;
    }
}
