package net.bytebuddy.test.packaging;

import net.bytebuddy.dynamic.scaffold.FieldLocatorForClassHierarchyTest;

public class FieldLocatorTestHelper extends FieldLocatorForClassHierarchyTest {
    /* empty */
}
